@@ -1,12 +1,12 @@
 from collections import defaultdict
+from .document_editor import DocumentEditor
 from typing import (Any, Callable, DefaultDict, Generator, List, Optional,
-                    Tuple, Dict, TypeVar)
+                    Tuple, Dict, TypeVar, Union)
 from PIL import Image, ImageDraw
 import numpy as np
 from math import pi
 import tatsu
 from tatsu.walkers import NodeWalker
-import json
 import sys
 import os
 
@@ -467,14 +467,15 @@ class ScopeStack:
                 return dic[symbol]
         return None
 
-    def put(self, symbol: str, value: Any) -> None:
+    def put(self, symbol: str, value: Any, min_layer: int = 0) -> None:
         assert self.stack
-        index = 0
 
-        while index < len(self.stack) - 1 and symbol not in self.stack[index]:
-            index += 1
+        for index in range(len(self.stack) - 1, min_layer - 1, -1):
+            if symbol in self.stack[index]:
+                self.stack[index][symbol] = value
+                return
 
-        self.stack[index][symbol] = value
+        self.stack[-1][symbol] = value
 
     def put_on_last(self, symbol: str, value: Any) -> None:
         assert self.stack
@@ -512,18 +513,30 @@ ARIT_OPS: Dict[str, Callable[[int, int], int]] = {
 
 MapperASTNode = Any
 
+MapperInbuiltFunction = Callable[['MapperWalker', List[Any]], Any]
+
+
+class MapperInbuiltFunctionWrapper():
+    def __init__(self, function: MapperInbuiltFunction):
+        self.function = function
+
 
 class MapperWalker(NodeWalker):
 
     def __init__(self, glob_scope: ScopeStack,
                  inbuilt: Optional[
-                     Dict[str, Callable[['MapperWalker', List[Any]], Any]]],
+                     Dict[str, Union[MapperInbuiltFunction, int]]],
                  ship: Ship,
                  mapp: Map) \
             -> None:
         self.ctx = glob_scope
         self.scope = glob_scope
-        self.inbuilt = inbuilt if inbuilt is not None else {}
+        if inbuilt:
+            for key, value in inbuilt.items():
+                if isinstance(value, int):
+                    self.scope.put(key, value)
+                else:
+                    self.scope.put(key, MapperInbuiltFunctionWrapper(value))
         self.ship = ship
         self.mapp = mapp
         self.max_out = 50
@@ -622,10 +635,11 @@ class MapperWalker(NodeWalker):
 
         walked_args = [self.walk(v) for v in node.args]
 
-        if node.name in self.inbuilt:
-            return self.inbuilt[node.name](self, walked_args)
-
         function = self.scope.lookup(node.name)
+
+        if isinstance(function, MapperInbuiltFunctionWrapper):
+            return function.function(self, walked_args)
+
         if function is None:
             raise MapperRuntimeError(
                 f"function '{node.name}' is not defined")
@@ -676,7 +690,7 @@ class MapperWalker(NodeWalker):
 
     def walk__assigment(self, node: MapperASTNode) -> MapperASTNode:
         value = self.walk(node.value)
-        self.scope.put(node.name, value)
+        self.scope.put(node.name, value, min_layer=1)
         return value
 
     def walk__operation(self, node: MapperASTNode) -> MapperASTNode:
@@ -726,13 +740,10 @@ def opt_range_str(rng: Optional[Tuple[Optional[int], Optional[int]]]) \
     return f'from {bottom} to {top}', True
 
 
-MapperBuiltInFunction = Callable[[MapperWalker, List[Any]], Any]
-
-
 def argument_check(name: str, number: Optional[Tuple[Optional[int],
                                                      Optional[int]]] = None) \
-        -> Callable[[MapperBuiltInFunction], MapperBuiltInFunction]:
-    def dec(func: MapperBuiltInFunction) -> MapperBuiltInFunction:
+        -> Callable[[MapperInbuiltFunction], MapperInbuiltFunction]:
+    def dec(func: MapperInbuiltFunction) -> MapperInbuiltFunction:
         def wrapper(ctx: MapperWalker, args: List[Any]) -> Any:
             if not in_opt_range(len(args), number):
                 rng_string, rng_pl = opt_range_str(number)
@@ -785,7 +796,15 @@ INBUILT = {
     'rot': rot,
     'put': put,
     'pen': pen,
-    'col': col
+    'col': col,
+    'BLACK': 0,
+    'RED': 1,
+    'GREEN': 2,
+    'BLUE': 3,
+    'RIGHT': 1,
+    'LEFT': -1,
+    'TRUE': 1,
+    'FALSE': 0
 }
 
 Model = Any
@@ -838,7 +857,8 @@ def run_models(models: List[Model]) -> Tuple[Any, Map, Ship, List[str]]:
     return result, mapper_map, ship, walker.out
 
 
-def generate_level(folder: str, src: str, name: str) -> None:
+def generate_level(folder: str, src: str, name: str,
+                   prnt_out: bool = False) -> None:
 
     src = os.path.join(folder, src)
     org_name = name
@@ -853,7 +873,8 @@ def generate_level(folder: str, src: str, name: str) -> None:
 
     models = [compile_code(code)]
     _, mapp, ship, out = run_models(models)
-    print(out)
+    if prnt_out:
+        print(out)
 
     tiles = export_map(mapp, sep=';')
     with open(f"{name}.txt", "w") as f:
@@ -885,17 +906,17 @@ def save_json(path_name: str, org_name: str, test: str) -> None:
         with open(md_file, 'w') as f:
             f.write('!\n')
 
-    json_str = json.dumps({
+    json_base = f"{path_name}_base.json"
+    DocumentEditor().inject({
         'code': org_name,
         'title': org_name,
         'desc': desc,
         'pipeline': 'mapper',
         'tests': [test],
         'prereqs': prereqs
-    }, indent=4)
-
-    with open(filepath, "w") as f:
-        f.write(json_str)
+    }).inject_json_file(
+        json_base, optional=True
+    ).save(filepath)
 
 
 if __name__ == "__main__":
